/* GStreamer
 * Copyright (C) <2004> Jeremy Simon <jsimon13@yahoo.fr>
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>
#include <gst/gst.h>
#include <gst/tag/tag.h>

#include "gstmonkeysdec.h"

static GstPadTemplate *src_template, *sink_template;

GstElementDetails monkeysdec_details = {
  "MonkeysAudio decoder",
  "Codec/Decoder/Audio",
  "Decode MonkeysAudio audio stream (.ape)",
  "Jeremy SIMON <jsimon13@yahoo.fr> (C) 2002 Matthew T. Ashland",
};


/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  ARG_0
};

static void     gst_monkeysdec_base_init       (gpointer g_class);
static void	gst_monkeysdec_class_init      (GstMonkeysDecClass *klass);
static void	gst_monkeysdec_init	       (GstMonkeysDec *monkeysdec);
static void     gst_monkeysdec_loop            (GstElement *element);
static gboolean gst_monkeysdec_src_event       (GstPad *pad, GstEvent *event);
static gboolean gst_monkeysdec_src_query       (GstPad *pad, GstQueryType type,
		                                GstFormat *format, gint64 *value);
static GstPadLinkReturn
                gst_monkeysdec_srcconnect      (GstPad *pad, const GstCaps *caps);
static const GstQueryType*
                gst_monkeysdec_get_query_types (GstPad *pad);
static const GstFormat*
                gst_monkeysdec_get_formats     (GstPad *pad);
static GstElementStateReturn
        	gst_monkeysdec_change_state    (GstElement *element);
static void gst_monkeysdec_get_property        (GObject *object, guint prop_id, 
                                                GValue *value, GParamSpec *pspec);
static void gst_monkeysdec_set_property        (GObject *object, guint prop_id, 
                                                const GValue *value, GParamSpec *pspec);


static GstElementClass *parent_class = NULL;

GType gst_monkeysdec_get_type (void)
{
  static GType monkeysdec_type = 0;

  if (!monkeysdec_type) {
    static const GTypeInfo monkeysdec_info = {
      sizeof (GstMonkeysDecClass), 
      gst_monkeysdec_base_init,
      NULL,
      (GClassInitFunc) gst_monkeysdec_class_init,
      NULL,
      NULL,
      sizeof (GstMonkeysDec),
      0,
      (GInstanceInitFunc) gst_monkeysdec_init,
    };

    monkeysdec_type = g_type_register_static (GST_TYPE_ELEMENT, "GstMonkeysDec", &monkeysdec_info, (GTypeFlags)0);
  }
  return monkeysdec_type;
}


static GstCaps*
monkeys_caps_factory (void)
{
  return gst_caps_new_simple ("application/x-ape", NULL);
}

static GstCaps*
raw_caps_factory (void)
{
  return gst_caps_new_simple ("audio/x-raw-int",
      "endianness", 	G_TYPE_INT, G_BYTE_ORDER,
      "signed", 	G_TYPE_BOOLEAN, TRUE,
      "width", 		G_TYPE_INT, 16,
      "depth",    	G_TYPE_INT, 16,
      "rate",     	GST_TYPE_INT_RANGE, 11025, 48000,
      "channels", 	GST_TYPE_INT_RANGE, 1, 2,
      NULL);
}

static void
gst_monkeysdec_base_init (gpointer g_class)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (g_class);
  GstCaps *raw_caps, *monkeys_caps;

  raw_caps = raw_caps_factory ();
  monkeys_caps = monkeys_caps_factory ();

  sink_template = gst_pad_template_new ("sink", GST_PAD_SINK,
      GST_PAD_ALWAYS, monkeys_caps);
  src_template = gst_pad_template_new ("src", GST_PAD_SRC, 
      GST_PAD_ALWAYS, raw_caps);

  gst_element_class_add_pad_template (element_class, sink_template);
  gst_element_class_add_pad_template (element_class, src_template);
  gst_element_class_set_details (element_class, &monkeysdec_details);
}


static void
gst_monkeysdec_class_init (GstMonkeysDecClass *klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;


  gstelement_class = (GstElementClass*)klass;
  gobject_class = (GObjectClass*) klass;

  parent_class = GST_ELEMENT_CLASS (g_type_class_ref(GST_TYPE_ELEMENT));

  gstelement_class->change_state = gst_monkeysdec_change_state;
  
  gobject_class->get_property = gst_monkeysdec_get_property;
  gobject_class->set_property = gst_monkeysdec_set_property;
}


static GstPadLinkReturn
gst_monkeysdec_srcconnect (GstPad *pad, const GstCaps *caps)
{
  GstMonkeysDec *monkeysdec;

  monkeysdec = GST_MONKEYSDEC (gst_pad_get_parent (pad));

  if (monkeysdec->state == GST_MONKEYSDEC_STATE_INIT)
    return GST_PAD_LINK_DELAYED;

  return GST_PAD_LINK_OK;
}

static void
gst_monkeysdec_init (GstMonkeysDec * monkeysdec)
{
  monkeysdec->sinkpad = gst_pad_new_from_template (sink_template, "sink");
  gst_element_add_pad (GST_ELEMENT (monkeysdec), monkeysdec->sinkpad);

  monkeysdec->srcpad = gst_pad_new_from_template (src_template, "src");
  gst_element_add_pad (GST_ELEMENT (monkeysdec), monkeysdec->srcpad);
  gst_pad_set_link_function (monkeysdec->srcpad, gst_monkeysdec_srcconnect);

  gst_element_set_loop_function (GST_ELEMENT (monkeysdec), gst_monkeysdec_loop);

  gst_pad_set_event_function (monkeysdec->srcpad, (GstPadEventFunction)(gst_monkeysdec_src_event));
  gst_pad_set_query_function (monkeysdec->srcpad, gst_monkeysdec_src_query);
  gst_pad_set_query_type_function (monkeysdec->srcpad, (GstPadQueryTypeFunction)(gst_monkeysdec_get_query_types));
  gst_pad_set_formats_function (monkeysdec->srcpad, (GstPadFormatsFunction)(gst_monkeysdec_get_formats));

  gst_pad_use_explicit_caps (monkeysdec->srcpad);

  monkeysdec->state == GST_MONKEYSDEC_STATE_INIT;
}


static const GstFormat*
gst_monkeysdec_get_formats (GstPad *pad)
{
  static const GstFormat src_formats[] = {
/*    GST_FORMAT_BYTES,
    GST_FORMAT_DEFAULT,*/
    GST_FORMAT_TIME,
    (GstFormat)0
  };
  static const GstFormat sink_formats[] = {
    /*GST_FORMAT_BYTES,*/
    GST_FORMAT_TIME,
    (GstFormat)0
  };
  
  return (GST_PAD_IS_SRC (pad) ? src_formats : sink_formats);
}


static const GstQueryType*
gst_monkeysdec_get_query_types (GstPad *pad)
{
  static const GstQueryType gst_monkeysdec_src_query_types[] = {
    GST_QUERY_TOTAL,
    GST_QUERY_POSITION,
    (GstQueryType)0
  };
  
  return gst_monkeysdec_src_query_types;
}


static gboolean
gst_monkeysdec_src_query (GstPad *pad, GstQueryType type,
		          GstFormat *format, gint64 *value)
{
  gboolean res = TRUE;
  GstMonkeysDec *monkeysdec;

  monkeysdec = GST_MONKEYSDEC (gst_pad_get_parent (pad));

  if (monkeysdec->state == GST_MONKEYSDEC_STATE_INIT)
	  return FALSE;

  switch (type) {
    case GST_QUERY_TOTAL:
      switch (*format) {
        case GST_FORMAT_TIME:
            *value = monkeysdec->decomp->GetInfo (APE_INFO_LENGTH_MS) * 1000000LL;
            break;
        default:
            res = FALSE;
            break;
      }
      break;
    case GST_QUERY_POSITION:
      switch (*format) {
         default:
           *value = monkeysdec->decomp->GetInfo (APE_DECOMPRESS_CURRENT_MS)  * 1000000LL;
           break;
      }
    default:
      break;
  }

  return res;
}    


static gboolean
gst_monkeysdec_src_event (GstPad *pad, GstEvent *event)
{
  gboolean res = TRUE;
  GstMonkeysDec *monkeysdec; 

  monkeysdec = GST_MONKEYSDEC (gst_pad_get_parent (pad));

  switch (GST_EVENT_TYPE (event)) {
    /* the all-formats seek logic */
    case GST_EVENT_SEEK:
    {
      monkeysdec->seek_to = GST_EVENT_SEEK_OFFSET (event) * monkeysdec->frequency /  GST_SECOND;
      monkeysdec->state = GST_MONKEYSDEC_STATE_SEEK;
      break;
    }
    default:
      res = FALSE;
      break;
  }
  
  gst_event_unref (event);

  return res;
}


static void
gst_monkeysdec_ape_to_gst (GstTagList *list, const gchar *field, const gchar *value)
{
  if (strcasecmp (field, "title") == 0)
  {
    gst_tag_list_add (list, GST_TAG_MERGE_REPLACE,
                      GST_TAG_TITLE, value, NULL);
    return;
  }

  if (strcasecmp (field, "album") == 0)
  {
    gst_tag_list_add (list, GST_TAG_MERGE_REPLACE,
                      GST_TAG_ALBUM, value, NULL);
    return;
  }

  if (strcasecmp (field, "artist") == 0)
  {
    gst_tag_list_add (list, GST_TAG_MERGE_REPLACE,
                      GST_TAG_ARTIST, value, NULL);
    return;
  }

  if (strcasecmp (field, "year") == 0)
  {   
    GDate *date;
    guint year;

    year = strtoul (value, NULL, 10);
    date = g_date_new_dmy (1, G_DATE_JANUARY, year);
    
    year = g_date_get_julian (date);
    g_date_free (date);

    gst_tag_list_add (list, GST_TAG_MERGE_REPLACE,
                      GST_TAG_DATE, year, NULL);
    return;
  }

  if (strcasecmp (field, "comment") == 0)
  {
    gst_tag_list_add (list, GST_TAG_MERGE_REPLACE,
                      GST_TAG_COMMENT, value, NULL);
    return;
  }

  if (strcasecmp (field, "genre") == 0)
  {
    gst_tag_list_add (list, GST_TAG_MERGE_REPLACE,
                      GST_TAG_GENRE, value, NULL);
    return;
  }

  if (strcasecmp (field, "copyright") == 0)
  {
    gst_tag_list_add (list, GST_TAG_MERGE_REPLACE,
                      GST_TAG_COPYRIGHT, value, NULL);
    return;
  }

  if (strcasecmp (field, "composer") == 0)
  {
    gst_tag_list_add (list, GST_TAG_MERGE_REPLACE,
                      GST_TAG_PERFORMER, value, NULL);
    return;
  }

  if (strcasecmp (field, "isrc") == 0)
  {
    gst_tag_list_add (list, GST_TAG_MERGE_REPLACE,
                      GST_TAG_ISRC, value, NULL);
    return;
  }

  if (strcasecmp (field, "publicationright") == 0)
  {
    gst_tag_list_add (list, GST_TAG_MERGE_REPLACE,
                      GST_TAG_LICENSE, value, NULL);
    return;
  }

  if (strcasecmp (field, "track") == 0)
  {
    gchar *slash = NULL;
    guint track_number;
    guint total_tracks;

    slash = g_strrstr (value, "/");
    if (slash)
    {
      track_number = strtoul (value, NULL, 10);
      total_tracks = strtoul (slash + 1, NULL, 10);

      gst_tag_list_add (list, GST_TAG_MERGE_REPLACE,
                        GST_TAG_TRACK_COUNT, total_tracks, NULL);
    }
    else
      track_number = strtoul (value, NULL, 10);

    gst_tag_list_add (list, GST_TAG_MERGE_REPLACE,
                      GST_TAG_TRACK_NUMBER, track_number, NULL);
    return;
  }
}

static gboolean
gst_monkeysdec_update_metadata (GstMonkeysDec *monkeysdec)
{
  GstTagList *list = NULL;  
  gchar *name, *value;
  CAPETagField *field;
  CAPETag *tag = NULL;
  int CptField;

  tag = (CAPETag *) monkeysdec->decomp->GetInfo(APE_INFO_TAG);
  if (tag == NULL)
    g_return_val_if_fail (tag, FALSE);

  if (!tag->GetHasAPETag () && !tag->GetHasID3Tag ())
    return FALSE;
  
  list = gst_tag_list_new ();

  CptField = 0;
  while ((field = tag->GetTagField(CptField++)))
    if (field->GetFieldValueSize() != 0)
    {
      name = g_strndup (GetANSIFromUTF16(field->GetFieldName()), field->GetFieldSize());
      value = g_strndup (field->GetFieldValue(), field->GetFieldValueSize());

      gst_monkeysdec_ape_to_gst (list, name, value);
    }

  gst_element_found_tags_for_pad (GST_ELEMENT (monkeysdec), monkeysdec->srcpad, 0, list);

  delete tag;
  return TRUE;
}


static void
gst_monkeysdec_setup (GstMonkeysDec *monkeysdec)
{
  gint retval;

  monkeysdec->io = new sinkpad_CIO;
  monkeysdec->io->bs = gst_bytestream_new (monkeysdec->sinkpad);

  if (monkeysdec->io->bs == NULL)
  {
    GST_ELEMENT_ERROR (monkeysdec, RESOURCE, OPEN_READ, (NULL), ("Failed to initiliaze bytestream from sinkpad"));
    return;
  }    

  monkeysdec->io->sinkpad = monkeysdec->sinkpad;
  monkeysdec->io->Open ((wchar_t *)"");

  monkeysdec->decomp = CreateIAPEDecompressEx (monkeysdec->io, &retval);
  if (monkeysdec->decomp == NULL || retval != ERROR_SUCCESS)
  {
    GST_ELEMENT_ERROR (monkeysdec, LIBRARY, INIT, (NULL), ("Failed to initiliaze MonkeyAudio decoder engine : %d", retval));
    GST_ELEMENT_ERROR (monkeysdec, LIBRARY, INIT, (NULL), ("Failed to initiliaze MonkeyAudio decoder engine"));
    return;
  }

  gst_monkeysdec_update_metadata (monkeysdec);

  monkeysdec->channels  = monkeysdec->decomp->GetInfo (APE_INFO_CHANNELS);
  monkeysdec->frequency = monkeysdec->decomp->GetInfo (APE_INFO_SAMPLE_RATE);
  monkeysdec->depth     = monkeysdec->decomp->GetInfo (APE_INFO_BITS_PER_SAMPLE);

  if (!GST_PAD_CAPS (monkeysdec->srcpad)) {
    gst_pad_set_explicit_caps (monkeysdec->srcpad,
       gst_caps_new_simple ("audio/x-raw-int",
          "endianness",  G_TYPE_INT, G_BYTE_ORDER,
          "signed",      G_TYPE_BOOLEAN, TRUE,
          "width",       G_TYPE_INT, monkeysdec->depth,
          "depth",       G_TYPE_INT, monkeysdec->depth,
          "rate",        G_TYPE_INT, monkeysdec->frequency,
          "channels",    G_TYPE_INT, monkeysdec->channels,
          NULL));
  }

  monkeysdec->io->eos = FALSE;

  monkeysdec->total_samples = 0;
  monkeysdec->seek_to = 0;
  monkeysdec->state = GST_MONKEYSDEC_STATE_PLAY;
  monkeysdec->block_align = monkeysdec->decomp->GetInfo (APE_INFO_BLOCK_ALIGN);
}

static void
gst_monkeysdec_loop (GstElement *element)
{
  GstMonkeysDec *monkeysdec;
  gint blocks_retrieved;
  GstBuffer *buffer_out;
  GstFormat format;
  gint64 timestamp;

  g_return_if_fail (element != NULL);
  g_return_if_fail (GST_IS_MONKEYSDEC (element));
	
  monkeysdec = GST_MONKEYSDEC (element);

  if (!GST_PAD_IS_USABLE (monkeysdec->srcpad))
    return;

  if (monkeysdec->state == GST_MONKEYSDEC_STATE_INIT)
    gst_monkeysdec_setup (monkeysdec);

  if (monkeysdec->state == GST_MONKEYSDEC_STATE_SEEK)
  {
    monkeysdec->decomp->Seek (monkeysdec->seek_to);
    monkeysdec->seek_to = 0;

    monkeysdec->state = GST_MONKEYSDEC_STATE_DISCONT; 
  }

  if (monkeysdec->state == GST_MONKEYSDEC_STATE_DISCONT)
  {
    GstEvent *discont;

    format = GST_FORMAT_TIME;
    gst_monkeysdec_src_query (monkeysdec->srcpad, GST_QUERY_POSITION, &format, &timestamp);
    
    discont = gst_event_new_discontinuous (FALSE, GST_FORMAT_TIME, timestamp, NULL);
    gst_pad_push (monkeysdec->srcpad, GST_DATA (discont));

    monkeysdec->state = GST_MONKEYSDEC_STATE_PLAY;
  }
  
  blocks_retrieved = BLOCK_TO_RETRIEVE;
  buffer_out = gst_buffer_new_and_alloc (BLOCK_TO_RETRIEVE * monkeysdec->block_align);
  monkeysdec->decomp->GetData((char *) GST_BUFFER_DATA (buffer_out), BLOCK_TO_RETRIEVE, &blocks_retrieved);
  
  format = GST_FORMAT_TIME;
  gst_monkeysdec_src_query (monkeysdec->srcpad, GST_QUERY_POSITION, &format, &timestamp);

  GST_BUFFER_SIZE (buffer_out) = blocks_retrieved * monkeysdec->block_align;
  GST_BUFFER_TIMESTAMP (buffer_out) = timestamp;
  GST_BUFFER_DURATION  (buffer_out) = monkeysdec->block_align * blocks_retrieved * 
                                      GST_SECOND / monkeysdec->frequency;

  gst_pad_push (monkeysdec->srcpad, GST_DATA (buffer_out));

  if (blocks_retrieved != BLOCK_TO_RETRIEVE && monkeysdec->io->eos)
  {
    GstEvent *event;

    event = gst_event_new (GST_EVENT_EOS);
    gst_pad_push (monkeysdec->srcpad, GST_DATA (event));
    gst_element_set_eos (element);
  }
}



static GstElementStateReturn
gst_monkeysdec_change_state (GstElement *element)
{
  GstMonkeysDec *monkeysdec;

  monkeysdec = GST_MONKEYSDEC (element);

  switch (GST_STATE_TRANSITION (element)) {
    case GST_STATE_NULL_TO_READY:
      break;
    case GST_STATE_READY_TO_PAUSED:
      monkeysdec->state = GST_MONKEYSDEC_STATE_INIT;
      break;
    case GST_STATE_PAUSED_TO_PLAYING:
      break;
    case GST_STATE_PLAYING_TO_PAUSED:
      break;
    case GST_STATE_PAUSED_TO_READY:
      break;
    case GST_STATE_READY_TO_NULL:
      break;
  }

  parent_class->change_state (element);

  return GST_STATE_SUCCESS;
}

static void
gst_monkeysdec_set_property (GObject *object, guint prop_id, 
		             const GValue *value, GParamSpec *pspec)
{
  GstMonkeysDec *monkeysdec;
	      
  g_return_if_fail (GST_IS_MONKEYSDEC (object));

  monkeysdec = GST_MONKEYSDEC (object);

  switch (prop_id) {
    default:
      g_warning ("Unknown property id\n");
  }
}

static void 
gst_monkeysdec_get_property (GObject *object, guint prop_id, 
		             GValue *value, GParamSpec *pspec)
{
  GstMonkeysDec *monkeysdec;
	      
  g_return_if_fail (GST_IS_MONKEYSDEC (object));

  monkeysdec = GST_MONKEYSDEC (object);

  switch (prop_id) {
    default:
      g_warning ("Unknown property id\n");
  }
}
